"exchange": {
    "name": "{{ exchange_name | lower }}",
    "key": "{{ exchange_key }}",
    "secret": "{{ exchange_secret }}",
    "ccxt_config": {},
    "ccxt_async_config": {},
    "pair_whitelist": [
    ],
    "pair_blacklist": [
        "HT/.*"
    ]
}
